# Nombre d'espcies de dinosaures per famlia observades durant tres perodes
# del Cretaci a Dakota del Nord i Montana (Sheehan, P.M., Fastovsky, D.E.,
# Hoffmann, R.G., Berghaus, C.B., Gabriel, D.L. (1991). Sudden extinction of
# the dinosaurs: Latest Cretaceous, Upper Great Plains, U.S.A. Science, 254,
# 835-839)

# Objectes que ara hi ha al nostre espai de treball:
ls()
# O, equivalentment:
objects()

dinos <- read.table("Dinos.txt", header = TRUE)
dinos
class(dinos)

# Objectes que ara hi ha al nostre espai de treball:
ls()

# Petit exercici al marge:
# Hi hauria una important diferncia si fessim servir el fitxer 'Dinos2.txt'
# (Busqueu l'nica diferncia entre 'Dinos.txt' i 'Dinos2.txt'!!!)
dinos2 <- read.table("Dinos2.txt", header = TRUE)
dinos2
class(dinos2)

# Per veure la diferncia pregunteu quantes columnes tenen respectivament:
ncol(dinos)
ncol(dinos2)

rownames(dinos)
rownames(dinos2)
# Conclusi?

fix(dinos)
fix(dinos2)

# Recordem com accedir a elements o parts d'una matriu o data.frame
# Perode 2 ("middle") de la tercera famlia ("Pachycephalo"):
dinos[3,2]  # s a dir: fila 3, columna 2
# o b, ms intuitiu per una mica ms ineficient:
dinos["Pachycephalo","middle"]
# o b:
dinos[3, "middle"]   # etc...

# Nombre d'espcies per famlia observades al perode superior:
dinos[,"upper"]
# Equivalent a:
dinos[,1]
# (Hem accedit a la columna de nom "upper" del 'data.frame' 'dinos'
# o, equivalentment, hem accedit a la seva primera columna.
#
# Similarment pels altres perodes, per exemple 'dinos[,3]', etc.

# Nombre d'espcies de la famlia "Tyranno" pels tres perodes:
dinos[4,]
# O, ms intuitiu:
dinos["Tyranno",]

# Un 'data.frame' i una matriu 'matrix' essencialment representen la mateixa
# estructura de dades: informaci organitzada en files i columnes.
# Externament, la principal diferncia est en que els elements d'una matriu
# han de ser tots de la mateixa classe (p.e. tots numrics) mentre que en un
# 'data.frame' cada columna pot ser de classe diferent.
class(dinos)
class(dinos2)

class(dinos2[,1])
class(dinos2[,2])

# Si tots elements d'una taula de dades sn de la mateixa classe, s ms
# eficient (en general) representar-la com una 'matrix':
dinos.mat <- as.matrix(dinos)
dinos.mat
class(dinos.mat)

# Aparentment, tot segueix igual:
# Perode 2 ("middle") de la tercera famlia ("Pachycephalo"):
dinos.mat[3,2]
# o b:
dinos.mat["Pachycephalo","middle"]
# o b:
dinos.mat[3, "middle"]   

dinos.mat["Tyranno",]
# etc...

# Per atenci:
dinos2.mat <- as.matrix(dinos2)
dinos2.mat
class(dinos2.mat[,1])
class(dinos2.mat[,2])
# Per uniformitzar la classe de tots els elements de la taula, els ha convertit
# en informaci alfanumrica 'character'. s un tipus d'informaci molt poc
# adequada per les operacions que farem a continuaci, per exemple calcular
# logaritmes:
log(dinos2.mat[3,2])
log(dinos.mat[3,2])

ls()

# Eliminem els objectes 'dinos2...':
rm(dinos2)
rm(dinos2.mat)

ls()

# Si la biodiversitat (en termes de nombre d'espcies per famlia) havia anat
# decreixent al llarg dels tres perodes, possiblement es podria considerar
# un indici que els dinosaures estaven immersos en una "crisi" que ja venia
# de lluny. La confrontaci d'aquesta possibilitat davant la hiptesi que
# la biodiversitat s'havia mantingut estable, requeriria mesurar la diversitat
# mitjanant un ndex objectiu, com ara l'ndex de Shannon - Margalef

# Calcularem aquest ndex per mesurar la biodiversitat al perode "upper"
# Magatzemem aquestes freqncies en un vector (una nova variable, un nou objecte)
# al que anomenarem 'up.cretacic':
up.cretacic = dinos[,1]

# Aquestes variables contenen informaci en termes de freqncies absolutes.
# Per calcular l'ndex de Shannon ens calen freqncies relatives.
# El reconvertim en un vector de freqncies relatives:
up.cretacic <- up.cretacic / sum(up.cretacic)
up.cretacic
# Ara sumen 1:
sum(up.cretacic)

# Logaritme en base 2 de cada freqncia relativa:
log2(up.cretacic)
# Feu ?log per tenir informaci sobre diverses funcions logartmiques

# Vector format pel producte de cada freqncia pel seu propi logaritme:
up.cretacic * log2(up.cretacic)

# Tots aquests valors sumats i finalment la suma canviada de signe correspon
# al valor de l'ndex de Shannon
-sum(up.cretacic * log2(up.cretacic))

# En resum, si tinc tot un vector de freqncies absolutes (com era 'up.cretacic'
# al principi) i vull calcular sobre ell l'ndex de Shannon, he d'executar una
# seqncia d'instruccions o comandes R tal com:
up.cretacic <- up.cretacic / sum(up.cretacic)
-sum(up.cretacic * log2(up.cretacic))

# Per calcular el mateix ndex pels altres perodes hauria de fer una
# cosa similar:
mid.cretacic <- dinos[,"middle"]

mid.cretacic <- mid.cretacic / sum(mid.cretacic)
-sum(mid.cretacic * log2(mid.cretacic))

low.cretacic <- dinos[,"lower"]

low.cretacic <- low.cretacic / sum(low.cretacic)
-sum(low.cretacic * log2(low.cretacic))

# Com un efecte collateral de la nostra feina, ara hi ha uns quants objectes
# ms al nostre espai de treball:
ls()

# Anar calculant la biodiversitat aix s poc prctic, especialment si ho volem
# fer amb unes dades similars a aquestes per amb moltes ms columnes

# s molt ms operatiu i elegant crear una "funci" que de forma general calculi
# l'ndex de Shannon.

# Primer la crearem i la farem servir. Desprs analitzarem qu vol dir all que
# hem fet.

# A la funci que anem a crear l'anomenarem 'shannon'.
# Creaci de la funci 'shannon':
shannon <- function(freq.abs)
{
  freq.rel <- freq.abs / sum(freq.abs)
  return(-sum(freq.rel * log2(freq.rel)))
}

# Ara hi ha un objecte ms (la funci que acabem de crear)
# al nostre espai de treball:
ls()

# Utilitzaci de la funci 'shannon':
up.cretacic <- dinos[,1]
shannon(up.cretacic)
# o, directament:
shannon(dinos[,1])
# instruccions que tindrien el mateix resultat que:
shannon(c(50, 29, 3, 3, 4, 1))

# Similarment:
shannon(dinos[,2])
shannon(dinos[,3])

# Amb unes altres dades (p.e. totalment fictcies que ens inventem sobre la marxa):
shannon(c(30,30))
shannon(c(15,45))
shannon(seq(from=10, to=50, by=10))
shannon(rep(30,5))

# Qu hem fet quan hem creat la funci?

# Fixem-nos que hem utilitzat l'operador d'assignaci '<-' (o '=')
# com si cressim i li donssim valor a una variable que representaria unes
# dades (p.e. x <- 15, etc). El que passa s que ara aquestes "dades" sn
# una informaci que servir per a processar altres dades. Sn unes dades d'una
# altra "classe", dades per a processar altres dades:
class(dinos)
class(up.cretacic)
class(shannon)
# Com l'interpret de R reconeix que estem creant una funci?
# Per que tot just desprs de '<-' hi ha la paraula clau 'function'.
# A continuaci, entre parntesis, hi ha d'anar la "llista d'arguments"
# (o "els parmetres") de la funci, al nostre cas un nic argument, 'freq.abs',
# (recodeu: shannon <- function(freq.abs)), i a continuaci, entre els smbols
# '{' i '}', totes les instruccions que realment fan el processament de dades
# propi de la funci.
#
# En resum:
# nom de la funci <- function(parmetre1, parmetre2, etc) {instruccions que
# especifiquen com aquesta funci processa la informaci proporcionada pels parmetres}

# Al nostre cas recordeu que era:
shannon <- function(freq.abs)
{
  freq.rel <- freq.abs / sum(freq.abs)
  -sum(freq.rel * log2(freq.rel))
}

# Qu succeeix quant des de l'entorn de treball de R escribim una cosa com
# 'shannon(dinos[,3])'?

# Ens en podem fer una idea adequada fent uns petits experiments...
# En primer lloc, afegim unes quantes instruccions 'print' a la funci:
shannon <- function(freq.abs)
{
  print("variables presents ara, al principi d'executar 'shannon'")
  print(ls()) # llistat d'objectes presents al nostre entorn de treball
  print("valor de la variable freq.abs:")
  print(freq.abs)
  freq.rel <- freq.abs / sum(freq.abs)
  print("variables presents ara, un cop dins 'shannon' s'ha creat 'freq.rel'")
  print(ls()) # llistat d'objectes presents al nostre entorn de treball
  print("valor de la variable freq.rel:")
  print(freq.rel)
  -sum(freq.rel * log2(freq.rel))
}

# Ara executem aquestes instruccions i provem d'entendre que va passant:

# Objectes presents a l'entorn de treball R (all que en diuen "nivell de sessi"):
ls()
# En qualsevol moment, quan estem a nivell de sessi, podem accedir al contingut
# real de qualsevol d'aquests objectes (o modificar-lo), per exemple:
dinos
up.cretacic

# Executem la funci 'shannon':
shannon(dinos[,3])

# Qu ha passat?
# Primerament, s'ha avaluat l'expressi que hi havia entre parntesis desprs del nom de la funci
# (b, en realitat R t una caracterstica anomenada "avaluaci diferida" que faria que aix
# s'hagus de matisar una mica, per l'afirmaci anterior s essencialment correcta)
# Recordem que el valor de 'dinos[,3]' s el vector amb valors 50 29  3  3  4  1.
# El pas anterior determina quina informaci li passem a la funci, quins parmetres o arguments.
# 
# A partir d'aquest moment, l'execuci de R entra en un altre entorn. Abandonem el nivell
# de sessi i entrem en l'entorn de l'execuci de la funci. Dins d'aquest nou entorn,
# en principi no es pot pressuposar que els objectes als que tenem accs a nivell de sessi
# estiguin disponibles. La via natural per a passar informaci d'entrada a una funci sn
# els seus arguments. L'argument 'freq.abs' s una variable que es crea dins l'entorn d'execuci
# de la funci i que d'entrada aquesta vegada concreta rep els valors 50 29  3  3  4  1.
# Amb aquesta variable es fan les operacions indicades i, a partir de la sentncia 
# freq.rel <- freq.abs / sum(freq.abs)
# es crea la nova variable 'freq.rel' dins l'entorn d'execuci de la funci, tal com el nou ls()
# posa de manifest.
# Si no hi ha res que ho modifiqui (ja en parlarem) les instruccions es van executant en l'ordre que
# estan escrites, l'ltima que s'executa determina el valor que finalment retorna la funci.
# En sortir de l'execuci de la funci, tots els objectes que s'han creat dins el seu entorn
# despareixen de memria:
ls()

# Tornem a la versi original de la funci:
shannon <- function(freq.abs)
{
  freq.rel <- dinos[,3] / sum(dinos[,3])
  -sum(freq.rel * log2(freq.rel))
}

# Anem a millorar una mica la funci anterior i aix de passada coneixerem unes
# quantes instruccions importants.

# Provem la funci pel cas que una de les freqncies s zero:
ambZeros <- up.cretacic
ambZeros[5] <- 0
shannon(ambZeros)
# Els resultat s NaN ("Not a Number") ats que logaritme de 0 s una
# indefinici, - infinit
log2(0)

# Encara que el modus d'operaci ms recomanable de R es basa en operacions
# en bloc, per vectors sencers, si utilitzem la instrucci 'for' tindrem ms
# control sobre els elements particulars (al cap i a la fi, el problema s que
# un d'ells s nul).
# Aquesta nova versi de la funci fa el mateix (incls l'error, que encara no
# hem corregit):

shannon <- function(freq.abs) {
  freq.rel <- freq.abs / sum(freq.abs)
  suma <- 0
  for (i in 1:length(freq.rel)) {
      valor <- freq.rel[i] * log2(freq.rel[i])
      suma <- suma + valor
  }
  return(-suma)
}

shannon(up.cretacic)
shannon(ambZeros)

# La instrucci 'for' t la segent sintaxi: for (variable in vector) instrucci
#
# 'variable' (una variable que es crea sobre la marxa per executar la iteraci,
# un cop finalitzada desapareix de la memria) anir prenent successivament tots
# els valors de 'vector' i per cada un d'aquests valors s'executar 'instrucci'.
# 'instrucci' pot ser una simple instrucci R o b una instrucci composta, tot
# un grup d'instruccions R emmarcades entre {}. Per exemple, si volem evitar
# la indeterminaci log2(0), podem escriure:
shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	suma = 0
	for (i in 1:length(freq.rels))
	{
		if (freq.rels[i] <= 0) next
		suma = suma + freq.rels[i] * log2(freq.rels[i])
	}
	- suma
}

shannon(up.cretacic)
shannon(ambZeros)

# A la versi anterior de la funci 'shannon' hem fet servir dues coses noves:
# les sentncies 'if' i 'next'.
# Una de les versions de la sentncia 'if' t la sintaxi: if (expressi booleana) sentncia
# Si 'expressi booleana' resulta finalment en un valor TRUE, s'executa 'sentncia',
# en cas contrari (valor FALSE) 'sentncia' s'ignora i es continua amb la
# instrucci segent.
# Com sempre, 'sentncia' pot ser una instrucci simple o tot un grup emmarcat
# en {}, que llavors actua com una unitat (o s'executen totes les instruccions
# entre {} o s'ignoren totes.
# 'next' provoca que la iteraci concreta del 'for' (o de qualsevol altre
# instrucci iterativa com ara 'while', veieu ms endavant) es suspengui i
# s'inici automticament la segent iteraci. Per exemple, si l'ndex del 'for'
# a la funci 'shannon' val i=5, en haver fet 'shannon(ambZeros)' tindrem que
# 'freq.rels[i] == 0' valdr TRUE, l'ordre 'next' que hi ha a continuaci
# s'executar i per tant la iteraci aquella es suspendr (per tant, no
# s'executar 'suma = suma + freq.rels[i] * log2(freq.rels[i])'), i passar a
# valdre 6 i es continuar amb el procs.
#
# Fixeu-vos tamb que per decidir si dues coses sn iguals hem utilitzat
# l'operador lgic d'igualtat '=='. 'a==b' t com a resultat TRUE si el
# valor d'a coincideix amb el de b, i FALSE en cas contrari (sense alterar
# els seus respectius valors), mentre que 'a=b' (que equivaldria a 'a<-b') s
# una "assignaci", substitueix el valor que tenia 'a' pel que t 'b' (o si
# l'objecte 'a' no existia, el crea amb valor inicial igual al de 'b').

# Una altra modalitat de la sentncia 'if' s:
# if (expressi booleana) sentncia1 else sentencia2
# on novament 'sentncia1' i 'sentncia2' poden ser instruccions simples o
# grups {}. Si 'expressi booleana' val TRUE s'executa 'sentncia1' i
# 'sentncia2' s'ignora. En cas contrai s'ignora 'sentncia1' i s'executa
# 'sentncia2'.
# Utilitzant aquesta altra possibilitat de 'if', podrem escriure:

shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	suma = 0
	for (i in 1:length(freq.rels))
	{
		if (freq.rels[i] == 0)
		    suma = suma + 0
    else
		    suma = suma + freq.rels[i] * log2(freq.rels[i])
	}
	- suma
}


shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	suma = 0
	for (i in 1:length(freq.rels))
	{
		if (freq.rels[i] < 0)
		    stop("No hi pot haver frequencies negatives")
    else {
        if (freq.rels[i] == 0)
		        suma = suma + 0
        else
		        suma = suma + freq.rels[i] * log2(freq.rels[i])
    }
	}
	- suma
}

shannon(up.cretacic)
shannon(ambZeros)

# En realitat 'if ... else ...' s una unitat per si mateixa, de fet els {}
# caldrien (encara que molts programadors prefereixen posar-los sempre, amb la
# finalitat que el codi sigui ms clar, p.e. per deixar clar quines instruccions
# fa repetir un 'for')
shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	suma = 0
	for (i in 1:length(freq.rels))
		if (freq.rels[i] == 0)
		    suma = suma + 0
    else
		    suma = suma + freq.rels[i] * log2(freq.rels[i])
	- suma
}

shannon(up.cretacic)
shannon(ambZeros)

# Igualment, suma = suma + 0 no fa res, tamb seria vlid:
shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	suma = 0
	for (i in 1:length(freq.rels))
		if (freq.rels[i] == 0)
		{
    }else
		    suma = suma + freq.rels[i] * log2(freq.rels[i])
	- suma
}

shannon(up.cretacic)
shannon(ambZeros)

# Fixeu-vos en el grup {} buid. Ara cal indicar que l'acci associada a 'if' s
# no res, {}, 'else' no pot anar immediatament a continuaci de 'if (...)'.

# Finalment, la versi ms concisa d'aquesta idea seria ('!=' vol dir "no igual"):
shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	suma = 0
	for (i in 1:length(freq.rels))
		if (freq.rels[i] != 0)
		    suma = suma + freq.rels[i] * log2(freq.rels[i])
	- suma
}

shannon(up.cretacic)
shannon(ambZeros)

# Les instruccions 'for', 'if' i 'next' que hem fet servir abans sn un exemple
# d'instruccions de control de flux d'execuci, que d'alguna manera trenquen
# l'ordre d'execuci seqencial. Aquest tipus d'instruccions existeixen (amb
# diverses variants) a tots els llenguatges de programaci i sn esencials
# per a poder expressar algorismes. Feu:
?"for"
# per a obtenir ajuda sobre les instruccions de control de flux disponibles en
# R. Com a complement d'aquesta ajuda, illustrem com es podrien utilitzar les
# sentncies 'while' i 'repeat' per a crear versions alternatives de la funci
# shannon:

shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	suma = 0
	i = 1
  k = length(freq.rels)
	while (i <= k) {
		if (freq.rels[i] != 0)
		    suma = suma + freq.rels[i] * log2(freq.rels[i])
    i = i + 1
  }
	return(- suma)
}

shannon(up.cretacic)
shannon(ambZeros)

shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	suma = 0
	i = 1
  k = length(freq.rels)
	repeat {
    if (i > k) break
    if (freq.rels[i] != 0)
		    suma = suma + freq.rels[i] * log2(freq.rels[i])
    i = i + 1
  }
	return(- suma)
}

shannon(up.cretacic)
shannon(ambZeros)


# Malgrat que les instruccions com 'for' sn essencials, l'estil ms genu de R,
# i normalment el ms eficient, es basa en la utilitzaci de funcions que
# operen en bloc sobre vectors i estructures similars.
# Al present exemple, una funci molt adequada seria 'ifelse'. Aquesta funci
# espera rebre tres arguments:
# ifelse(test, yes, no)
# on 'test' correspon a un vector boole (un vector de valors TRUE o FALSE) i
# 'yes' i 'no' corresponen a vectors de la mateixa longitud que 'test'.
# El resutat retornat s tamb un vector de la mateixa longitud que els anteriors.
# A les posicions on 'test' val TRUE, el valor retornat s el de la corresponent
# posici de 'yes' i a les posicions on 'test' val FALSE, al resultat hi ha el
# valor corresponent de 'no'.

# Per exemple:
up.cretacic > 0
# s un vector boole amb la mateixa longitud que 'up.cretacic'.
# Comproveu el resultat doncs de:
ifelse(up.cretacic > 0, up.cretacic * log2(up.cretacic), 0)

# Per tant, una versi molt concisa de la funci 'shannon' seria:

shannon <- function(freq.abs)
{
	freq.rels = freq.abs / sum(freq.abs)
	- sum(ifelse(freq.rels > 0, freq.rels * log2(freq.rels), 0))
}

shannon(up.cretacic)
shannon(ambZeros)

shannon(up.cretacic)
hcharvat(up.cretacic)
renyi(up.cretacic)
hcharvat(up.cretacic, 3)
hcharvat(alpha = 3, freq.abs = up.cretacic)
hcharvat(up.cretacic, alpha = 3)

# Una funci que calcula totes 3 diversitats sobre unes mateixes dades:
divs <- all.divers(up.cretacic, alpha = 3)
divs
divs[3]
divs["Renyi"]

# Apliquem 'shannon' a totes les columnes de dinos:
# creaci d'un vector numric de la llargada adequada
biodivShannon <- numeric(ncol(dinos))   
# a cada element de 'biodivShannon' li assigno el clcul de la diversitat    
for (i in 1:length(biodivShannon)) biodivShannon[i] <- shannon(dinos[,i])
# Pega: no t noms, em puc oblidar de a qu corresponen aquests valors
biodivShannon

# (o igualment hcharvat, o renyi:)
biodivRenyi <- numeric(ncol(dinos)) 
for (i in 1:ncol(dinos)) biodivRenyi[i] <- renyi(dinos[,i], alpha = 1.5)
names(biodivRenyi) <- colnames(dinos)
biodivRenyi

# Una manera molt ms concisa:
apply(dinos, 2, shannon)

# Explicaci de la funci 'apply':
# Per la matrix o data.frame 'dinos', al llarg de la seva segona dimensi (per
# cada columna), aplica la funci 'shannon'

# Altres exemples d'utilitzaci d'apply:
# Suma (freqncia total) per files:
apply(dinos, 1, sum)
# Suma per columnes:
apply(dinos, 2, sum)

# Com veieu, un avantatge de 'apply' s que si la matriu de partida t noms
# de fila o de columna, s'aprofiten al resultat

# Entropia de Havrda-Charvat amb alpha = 1.5 per cada perode:
apply(dinos, 2, hcharvat, alpha = 1.5)
apply(dinos, 2, hcharvat, 1.5)

# Cada columna (o fila) de la matriu 'dinos' s'utilitza com el primer argument
# de la funci aplicada (al darrer exemple, 'hcharvat').
# Si la funci que apliquem a cada columna (o fila) t un segon, tercer, ...
# argument (p.e. 'hcharvat' t un segon argument, 'alpha'), aquests arguments
# es poden proporcionar al final: apply(...., alpha = 1.5)

# Lgicament
apply(dinos, 2, hcharvat)
# implica utilitzar el valor per defecte de alpha: alpha = 2

# Per matrius (arrays) de ms de 2 dimensions, 'apply' tamb s adequada:
# Creem una hipermatriu tridimensional de 2 files, 3 columnes i 4 hipercolumnes (?):
hiperMatrix <- array(1:24, dim=c(2,3,4))
hiperMatrix
# sumem al llarg de la tercera dimensi:
apply(hiperMatrix, 3, sum)
# sumem respecte de les dues primeres dimensions:
apply(hiperMatrix, c(1,2), sum)
# A partir de l'observaci dels valors inicialment a 'hiperMatrix', intenteu
# veure clar qu ha sumat i aix acabareu d'entendre com funciona 'apply'

# Amb 'apply' podrem crear fcilment una taula amb cada ndex de biodiversitat
# per cada perode:
apply(dinos, 2, all.divers)
apply(dinos, 2, all.divers, alpha = 3)

# *************** La famlia 'apply' de funcions *********************
# La funci anterior forma part de tota una srie de funcions que repetidament
# executen una altra funci per tota una estructura de dades.
# En general, utilitzar aquestes funcions s ms eficient que fer la mateixa
# cosa utilitzant 'for', 'while' etc, encara que tamb sovint el codi queda
# ms crptic.

# ************************** Llistes *********************************
# Moltes de les funcions 'apply' actuen sobre una estructura de dades
# anomenada 'list'. Es pot entendre com un vector que admet elements de
# diferents classes, on cada element pot ser tant complicat com es vulgui
# (i en particular pot ser una altra llista).
# Tpica instrucci per crear llistes:

estudi.dinos <- list(
  dades = dinos, 
  referencia = "Hoffmannet et al. Science, 254, 835-839, 1991",
  resultats = apply(dinos, 2, all.divers)
)

# Ho podrem haver posat tot en una lnia:
estudi.dinos <- list(dades = dinos, referencia = "Hoffmannet et al. Science, 254, 835-839, 1991", resultats = apply(dinos, 2, all.divers))
# per crec que de la primera manera dna una idea ms clara que s una llista
# de 3 coses diferents: un 'data.frame' amb les dades originals, un valor
# de classe 'character' (text) amb la referncia bibliogrfica i una 'matrix'
# amb els resultats.

# Aparentment sembla un vector:
length(estudi.dinos)
names(estudi.dinos)
estudi.dinos[1]
estudi.dinos[2]
estudi.dinos[3]
# Per s una llista:
class(estudi.dinos)

# Diferncia entre '[' i '[['
# Amb l'operador '[' una part d'una llista s una altra llista:
estudi.dinos[1]
class(estudi.dinos[1])
# s una llista formada solament pel primer element de 'estudi.dinos'

# Una llista formada pel primer i tercer elements:
estudi.dinos[c(1,3)]
# O equivalentment:
estudi.dinos[-2] # tota la llista excepte el segon element

# En canvi:
estudi.dinos[[1]]
class(estudi.dinos[[1]])
# (compareu amb class(estudi.dinos[1])!)
# '[' torna una subllista mentre que '[[' torna l'element que realment hi ha all

# Aix s un error:
estudi.dinos[[-2]] # I lgicament tamb estudi.dinos[[c(1,3)]]
# Aquest error s molt comprensible: alg en sabria donar la ra?

# Una altra manera d'accedir als elements d'una llista: l'operador '$':
estudi.dinos$referencia

# Equival a:
estudi.dinos[[2]]
estudi.dinos[["referencia"]]

# Una llista es pot assimilar a l'estructura de carpetes d'un sistema operatiu,
# que poden contenir altres carpetes indefinidament de forma recurrent.
# De la mateixa manera que si fossin carpetes en Windows escriurem:
# estudi.dinos/referencia
# per llistes R fem:
# estudi.dinos$referencia (i en general llista$subllista1$subllista2... etc)

# Moltes funcions R retornen els seus resultats en forma de 'list'.
# Per exemple, la funci 'chisq.test' que calcula el conegut test khi-quadrat:

khi.upper <- chisq.test(dinos[,"upper"])

# Amb la instrucci anterior hem fet una prova khi-quadrat de bondat d'ajust
# de les freqncies del perode "upper" a les proporcions teriques
# c(1/6, 1/6, 1/6, 1/6, 1/6, 1/6) (feu ?chisq.test per ms detalls)

# El resultat s una llista bastant llarga.
# Per exemple, si volem saber el p-valor del test anterior:
khi.upper$p.value   # O khi.upper[["p.value"]], etc
# L'estadstic khi-quadrat de bondat d'ajust:
khi.upper$statistic  # Tindria sentit com a mesura de biodiversitat?
# Els graus de llibertat:
khi.upper$parameter
# Les freqncies esperades segons la hiptesi nulla (1/6, 1/6, ...):
khi.upper$expected
# etc.

class(khi.upper)
# No diu que sigui de classe 'list' sin 'htest'
# En realitat s que s 'list'. 'htest' s una "especialitzaci" o "descendent"
# de 'list'. Per per entendre b aquest conceptes haurem de fer un curs de
# programaci orientada a objectes en R.

# Una altra versi especialitzada de 'list' s la classe 'data.frame'.
# En realitat un 'data.frame' s una llista de columnes, cada columna s un vector
# (per aquesta ra, cada columna d'un 'data.frame' pot ser de classe diferent).
# Un 'data.frame' s una llista "tunejada" per que sembli que es comporta com
# una matriu, per en realitat conserva (hereda) els seus origens 'list'.
# Per aquesta ra haurem d'entendre que totes aquestes coses sn legals per
# algunes diferents entre si:
dinos$upper
dinos[["upper"]]
dinos["upper"]
dinos[,"upper"]
dinos["Tyranno","upper"]
# I tamb haurem d'entendre aix:
class(dinos[["upper"]])
class(dinos["upper"])
class(dinos[,"upper"])
class(dinos["Tyranno","upper"])

# Tornem a les funcions 'apply':
# 'lapply' (abreviatura de "list apply") recorre una llista,
# executant sobre la llista la funci que li indiquem:
# per exemple, si volem imprimir tots els elements de khi.upper:
lapply(khi.upper, print)

# Si volem calcular els tres ndexos de diversitat sobre cada columna de dinos:
lapply(dinos, all.divers)
# o lapply(dinos, all.divers, alpha = 3) si volgussim fer servir arguments
# addicionals (els posarem al final)

# Per defecte, el resultat de 'lapply' s una llista. En canvi:
sapply(dinos, all.divers)
# intenta simplificar al mxim el resultat. En aquest cas el pot convertir en 
# una matriu i s el que fa.
# El resultat seria el mateix que el de:
apply(dinos, 2, all.divers)
# encara que 'sapply' i 'lapply' tenen fama de ser ms rpides

# Altres funcions similars sn 'tapply', 'vapply', 'replicate'.
# Feu per exemple ?lapply per tenir informaci sobre funcions d'aquest tipus i
# les seves moltes possibilitats